use io;

/* Read instance data. The input files follow the "Kone" format */
function input() {
    local usage = "Usage: localsolver rcpsp_producer_consumer.lsp inFileName=instanceFile "
            + "[outFileName=outputFile] [lsTimeLimit=timeLimit]";
    if (inFileName == nil) throw usage;

    inFile = io.openRead(inFileName);
    nbTasks = inFile.readInt(); // Number of operations
    nbResources = inFile.readInt(); // Number of renewable resources
    nbInventories = inFile.readInt(); // Number of inventories
    capacity[0...nbResources] = inFile.readInt();
    initLevel[0...nbInventories] = inFile.readInt();

    for [i in 0...nbTasks] {
        duration[i] = inFile.readInt();
        weight[i][0...nbResources] = inFile.readInt();
        for [r in 0...nbInventories] {
            startCons[i][r] = inFile.readInt();
            endProd[i][r] = inFile.readInt(); 
        }
        nbSuccessors[i] = inFile.readInt();
        successors[i][0...nbSuccessors[i]] = inFile.readInt()-1;
    }
    
    inFile.close();
    
    finishLevel[r in 0...nbInventories] = initLevel[r];
    for [r in 0...nbInventories] {
        for [i in 0...nbTasks] {
            finishLevel[r] -= startCons[i][r];
            finishLevel[r] += endProd[i][r];
        }
    }
    
    timeHorizon = sum[i in 0...nbTasks](duration[i]);
}

/* Declare the optimization model */
function model() {
    tasks[i in 0...nbTasks] <- interval(0, timeHorizon);
    
    for [i in 0...nbTasks] {
        constraint length(tasks[i]) == duration[i];
        
        for [s in 0...nbSuccessors[i]] {
            constraint end(tasks[i]) <=  start(tasks[successors[i][s]]);
        }
    }
    
    makespan <- max[i in 0...nbTasks] (end(tasks[i]));
    
    for [r in 0...nbResources] {
        constraint and(0...makespan,
            t => (sum[i in 0...nbTasks](weight[i][r] * (contains(tasks[i],t))) <= capacity[r]));
    }
    
    for [r in 0...nbInventories] {
        constraint and(0...makespan,
            t => (0 <=  initLevel[r] +
                  sum[i in 0...nbTasks]((endProd[i][r] * (end(tasks[i]) <= t)) - 
                                      (startCons[i][r] * (start(tasks[i]) <= t)))));
    }
    
    minimize makespan;
}

/* Parameterize the solver */
function param() {
    if (lsTimeLimit == nil) lsTimeLimit = 60;
}

function output() {
    if (outFileName != nil) {
        outFile = io.openWrite(outFileName);
        println("Solution written in file ", outFileName);
        outFile.println(makespan.value);
        for [i in 0...nbTasks] {
            outFile.println(i + 1, " ", tasks[i].value.start, " ", tasks[i].value.end);
        }
    }
}
